# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp-foglamp-documentation.readthedocs-hosted.com/
# FOGLAMP_END

""" RTSP Stream Plugin
"""
__author__ = "Himanshu Vimal"
__copyright__ = "Copyright (c) 2023 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import asyncio
import copy
import logging
import os
import time
from threading import Thread
import cv2
import numpy as np

from foglamp.common import logger
from foglamp.plugins.common import utils
import async_ingest

from foglamp.plugins.south.rtsp.rtsputils import RtspUtils
from foglamp.plugins.south.rtsp.videostream import VideoStream
from foglamp.plugins.south.rtsp.rtsputils import RtspUtils

_LOGGER = logger.setup(__name__)

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Capture RTSP Stream',
        'type': 'string',
        'default': 'rtsp',
        'readonly': 'true'
    },
    'asset_name': {
        'description': 'Asset name',
        'type': 'string',
        'default': 'stream',
        'order': '1',
        'displayName': 'Asset Name'
    },
    'stream_url': {
        'description': 'The URL of the network stream.',
        'type': 'URL',
        'default': 'rtsp://ip:port',
        'order': '2',
        'displayName': 'Stream URL',
        'mandatory': 'true'
    },
    'frames_per_interval': {
        'description': 'Frames to capture in given interval.',
        'type': 'integer',
        'default': '10',
        'order': '3',
        'displayName': 'Frame per Interval',
        'mandatory': 'true'
    },
    'interval': {
        'description': 'Interval in seconds over which to collect frames',
        'type': 'integer',
        'default': '1',
        'order': '4',
        'displayName': 'Interval',
        'mandatory': 'true'
    },
    'resize_frame': {
        'description': 'Resize frames to resolution defined below',
        'type': 'boolean',
        'default': 'false',
        'order': '5',
        'displayName': 'Resize Frame',
        'group': 'Frame Settings'
    },
    'resize_width': {
        'description': 'The resized width of the captured frame.',
        'type': 'integer',
        'default': '640',
        'order': '6',
        'displayName': 'Resize Width',
        'validity': "resize_frame == \"true\"",
        'group': 'Frame Settings'
    },
    'resize_height': {
        'description': 'The resized height of the captured frame.',
        'type': 'integer',
        'default': '480',
        'order': '7',
        'displayName': 'Resize Height',
        'validity': "resize_frame == \"true\"",
        'group': 'Frame Settings'
    },
    'authentication': {
        'description': 'Authentication method to be used for RTSP stream',
        'type': 'enumeration',
        'default': 'None',
        'order': '8',
        'displayName': 'Authentication',
        'options': ['None', 'Basic'],
        'group': 'Authentication'
    },
    'username': {
        'description': 'Username to be used for authentication',
        'type': 'string',
        'default': 'user',
        'order': '9',
        'displayName': 'Username',
        'validity': "authentication == \"Basic\"",
        'group': 'Authentication'
    },
    'password': {
        'description': 'Password to be used for RTSP stream',
        'type': 'password',
        'default': 'password',
        'order': '10',
        'displayName': 'Password',
        'validity':"authentication == \"Basic\"",
        'group': 'Authentication'
    },
}

# # GLOBAL VARIABLES DECLARATION
c_callback = None
c_ingest_ref = None
loop = None
stream_processor = None

def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """

    return {
        'name': 'RTSP plugin',
        'version': '3.0.0-1-g4f782a5',
        'mode': 'async',
        'type': 'south',
        'interface': '2.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin with the merged config
        Args:
           config: JSON configuration document for the South plugin configuration category
        Returns:
           data: JSON object to be used in future calls to the plugin
        Raises:
    """
    data = copy.deepcopy(config)
    return data


def plugin_start(handle):
    """ Start the plugin operation - capture rtsp stream from URL and save readings
        Args:
            handle: handle returned by the plugin initialisation call
        Returns:
            Sends back the reading object (JSON doc) via C/python async ingest interface when it captures a
            video frame
        Raises:
    """
    global stream_processor, loop
    loop = asyncio.new_event_loop()
    try:
        def run():
            global loop
            loop.run_forever()

        global stream_processor
        stream_processor = StreamProcessor(handle)
        stream_processor.start()
    except Exception as ex:
        _LOGGER.exception("RTSP plugin failed to start. Details: %s", str(ex))
        raise
    else:
        _LOGGER.info("Plugin started")



def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
        it should be called when the configuration of the plugin is changed
        during the operation of the south service.
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """
    global stream_processor

    parameters_to_check = ['asset_name', 'stream_url', 'authentication',
                           'username', 'password']
    need_to_shutdown = RtspUtils.check_need_to_shutdown(handle, new_config, parameters_to_check)
    if not need_to_shutdown:
        stream_processor.handle_new_config(new_config)
        new_handle = plugin_init(new_config)    
    else:
        plugin_shutdown(handle)
        new_handle = plugin_init(new_config)
        plugin_start(new_handle)

    return new_handle

def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
        To be called prior to the south service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        None
    Raises:
        None
    """
    try:
        global stream_processor, loop
        if stream_processor:
            stream_processor.shutdown_in_progress = True
            stream_processor.videostream.stopped = True
            # allow the stream to stop
            time.sleep(3)
            # stopping every other thread one by one

            # checking if the thread is started or not.
            if stream_processor.thread_started:
                stream_processor.join()
            stream_processor = None

        if loop:
            loop.stop()
            loop = None

        _LOGGER.info('Plugin has shutdown')
    except Exception as ex:
        _LOGGER.exception(str(ex))
        raise


def plugin_register_ingest(handle, callback, ingest_ref):
    """ Required plugin interface component to communicate to South C server

    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C/Python async ingest interface
        ingest_ref: C opaque object required to passed back to C/Python async ingest interface
    """
    global c_callback, c_ingest_ref
    _LOGGER.debug("register ingest")
    c_callback = callback
    c_ingest_ref = ingest_ref

class StreamProcessor(Thread):
    def __init__(self, handle):
        super(StreamProcessor, self).__init__()

        # Get configuration
        self.asset_name = handle['asset_name']['value']
        self.stream_url = handle['stream_url']['value']
        self.frames_per_interval = int(handle['frames_per_interval']['value'])
        self.frame_interval = int(handle['interval']['value'])
        self.resize_frame = bool(handle['resize_frame']['value'] == 'true')
        self.shutdown_in_progress = False

        self.username = None
        self.password = None
        self.frame_height = None
        self.frame_width = None

        if self.resize_frame:
            self.frame_height = int(handle['resize_height']['value'])
            self.frame_width = int(handle['resize_width']['value'])
        
        if handle['authentication']['value'] == 'Basic':
            self.username = handle['username']['value']
            self.password = handle['password']['value']

        self.videostream = VideoStream(stream_url=self.stream_url,
                                                                username=self.username,
                                                                password=self.password)
        if self.videostream.initializeStream():
            # stream connected
            self.videostream.start()
        else :
            _LOGGER.error("Not able to capture stream. Check configuration and ensure stream is "
                          "running")
            # retry connection
            self.videostream.reconnect()
        
    def wait_for_frame(self):
        """ Waits for frame to become available else sleeps for 200 milliseconds.
                Args:
                       self-> a videostream object
                Returns: None
               Raises: None
           """
        while True:
            if self.videostream.frame is not None:
                return
            if self.shutdown_in_progress:
                return
            _LOGGER.debug("Waiting for frame")
            time.sleep(0.2)

    def handle_new_config(self, new_config):
        """
        If shutdown is not required then it changes the configuration on the fly.
        Args:
            new_config: The configuration during reconfigure.

        Returns:
              None
        """
        _LOGGER.debug("Handling the reconfigure without shutdown")
        self.resize_frame = bool(new_config['resize_frame']['value'] == 'true')
        self.frame_height = int(new_config['resize_height']['value'])
        self.frame_width = int(new_config['resize_width']['value'])
        self.frames_per_interval = int(new_config['frames_per_interval']['value'])
        self.frame_interval = int(new_config['interval']['value'])

    def run(self):        
        # wait for first frame to be read
        self.wait_for_frame()
        self.thread_started = True
        tick_frequency = cv2.getTickFrequency()
        while True:
            # check plugin shutdown
            if self.shutdown_in_progress:
                _LOGGER.info("Plugin shutdown initiated. Shutting down stream capture")
                break    

            reading_freq = self.frame_interval/self.frames_per_interval
            # Capture frame-by-frame
            global c_callback, c_ingest_ref
            t1 = cv2.getTickCount()
           
            # Taking the frame the stream
            frame1 = self.videostream.read()
            if frame1 is None:
                # waiting for steram to reconnect
                self.wait_for_frame()
                continue
                
            # BGR to RGB_LOGGER.debug("Waiting for frame")
            frame = frame1.copy()
            frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            if self.resize_frame:
                frame_rgb = cv2.resize(frame_rgb, (self.frame_width, self.frame_height))
            input_data = np.array(frame_rgb)

            # the readings to be inserted into the table
            objs = {}
            objs['frame'] = input_data
            data = {
                'asset': self.asset_name,
                'timestamp': utils.local_timestamp(),
                'readings': objs
            }    
            async_ingest.ingest_callback(c_callback, c_ingest_ref, data)                
            
            # wait as per frame capture interval
            time_elapsed = (cv2.getTickCount() - t1)/tick_frequency
            sleep_time = (reading_freq - time_elapsed)
            if sleep_time > 0:
                time.sleep(sleep_time)                   
            
        _LOGGER.debug("Stopping stream capture")
        self.videostream.stop()
        time.sleep(2)
