# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp-foglamp-documentation.readthedocs-hosted.com/
# FOGLAMP_END

""" Helper module for starting a separate thread for reading frame from the Camera Device
"""

import logging
from threading import Thread
import os
import time
import cv2

from foglamp.common import logger

_LOGGER = logger.setup(__name__)

class VideoStream:
    """ Camera object that controls video streaming from the Camera
    """

    def __init__(self, username=None, password=None, stream_url=None):     
        self.stopped = False
        self.frame = None
        self.grabbed = False
        self.stream_url = None
        # Initialize the camera image stream
        if stream_url:
            if username:
                index = stream_url.index('://')
                auth_str = username + ':' + password +'@'
                stream_url = stream_url[:index+3] + auth_str + stream_url[index+3:]
            self.stream_url = stream_url    

        if self.stream_url is None:
            _LOGGER.exception("Not able to find stream url. Please "
                                  "reconfigure the plugin and try again with a correct stream url")
        
    def initializeStream(self):
        # Open camera stream
        os.environ["OPENCV_FFMPEG_CAPTURE_OPTIONS"] = "rtsp_transport;udp"
        self.stream = cv2.VideoCapture(self.stream_url, cv2.CAP_FFMPEG)
        if not self.stream.isOpened():
            self.stream.release()
            self.stream = None
            return False
        _LOGGER.debug("Stream opened successfully at :{}".format(self.stream_url))
        return True

    def start(self):
        _LOGGER.info("Starting stream capture")
        # Start the thread that reads frames from the video stream
        t = Thread(target=self.update, args=(), name="Reader Thread")
        t.daemon = True
        t.start()

    def update(self):
        # Keep looping indefinitely until the thread is stopped
        while True:
            # If the camera is stopped, stop the thread
            if self.stopped:
                # Release camera resources
                self.stream.release()
                self.stream = None
                return

            # Otherwise, grab the next frame from the stream
            (self.grabbed, self.frame) = self.stream.read()
            if not self.grabbed:
                self.stream.release()
                _LOGGER.exception("Unable to capture frame, stream not functional")
                self.reconnect()
                return
            
    def read(self):
        # Return the most recent frame
        return self.frame

    def stop(self):
        # Indicate that the camera and thread should be stopped
        self.stopped = True
    
    
    def reconnect(self):
        _LOGGER.info("Retrying stream capture")
        rc = Thread(target=self.retryConnection, args=(), name="Reconnect Thread")    
        rc.daemon = True
        rc.start()

    def retryConnection(self):
        # Clean up resources and retry stream capture.
        is_stream_open = False
        while True:
            is_stream_open = self.initializeStream()
            if is_stream_open:
                _LOGGER.info("Camera stream re-connected")
                self.start()
                break

            # plugin shutdown
            if self.stopped:
                self.stream = None
                break

            # retry after 200 ms
            time.sleep(0.2)

            
                



